import time 
import numpy as np
import torch 
torch.manual_seed(42)
torch.set_num_threads(1)

from utils import gauge, trace
import pymanopt 
from pymanopt.manifolds import Product, Stiefel
from pymanopt.solvers import TrustRegions, ConjugateGradient

def hosrg(T, x):
    manifold = Stiefel(args.chi**2, args.chi, args.niter)
    
    def _cost(isometries, T):
        lnZ = 0.0
        for n in range(args.niter):
            f = T.abs().max()
            lnZ += 2**(-n)*torch.log(f)
            T = T / f

            U = isometries[n].view(args.chi, args.chi, args.chi)

            T = renormalize(T, U)
           
        lnZ += torch.log(trace(T))/2**args.niter
        return -lnZ

    @pymanopt.function.PyTorch
    def cost(isometries):
        return _cost(isometries, T)

    problem = pymanopt.Problem(manifold, cost)
    solver = TrustRegions()
    #solver = ConjugateGradient()
    xopt = solver.solve(problem, x=x)
    return xopt

def renormalize(T, U):
    T = torch.einsum('axob,amz,moyn,bnw->xwzy',(T, U, T, U))
    #TU1 = torch.tensordot(T, U, dims=([0], [0]))
    #TU2 = torch.tensordot(T, U, dims=([3], [1]))
    #T = torch.tensordot(TU1, TU2, dims=([1,2,3], [1,3,0]))
    #T = T.permute(0,3,1,2).contiguous()
    return T

def hotrg(T):
    #initialize isometry with hotrg
    lnZ = 0.0
    isometries = []
    for step in range(args.niter):
        f = T.abs().max()
        lnZ += 2**(-step)*torch.log(f)
        T = T / f

        #contract up-down
        vl, el = gauge(T, args.chi,'l')
        vr, er = gauge(T, args.chi,'r')

        U = vl if (el < er) else vr
        isometries.append(U.view(-1, U.shape[-1]))
        T = renormalize(T, U)
    lnZ += torch.log(trace(T))/2**args.niter
    print ('initial lnZ with hotrg', lnZ.item())
    return torch.stack(isometries).numpy()

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='')
    
    parser.add_argument("-beta", type=float, default=0.44068679350977147, help="beta")
    parser.add_argument("-D", type=int, default=2, help="D")
    parser.add_argument("-chi", type=int, default=24, help="chi")
    parser.add_argument("-niter", type=int, default=30, help="niter")
    parser.add_argument("-Nepochs", type=int, default=100, help="Nepochs")
    parser.add_argument("-init", default='random', choices=['random', 'hotrg'], help="initial isometry")
    parser.add_argument("-exact", type=float, default=-0.9296953983414709, help="exact free energy")
    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-cuda", type=int, default=-1, help="GPU #")
    args = parser.parse_args()

    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64
    print ('use', dtype)

    # initial tensor
    K = torch.tensor([args.beta], dtype=dtype, device=device)
    c = torch.sqrt(torch.cosh(K)/2.)
    s = torch.sqrt(torch.sinh(K)/2.)
    M = torch.stack([torch.cat([c+s, c-s]), torch.cat([c-s, c+s])])
    T = torch.einsum('ai,aj,ak,al->ijkl', (M, M, M, M)) 

    assert(args.chi in [2, 4, 16, 256]) # chi has to be power of 2 
    
    #first do a few steps of exact RG to grow bond dimension to args.chi
    #lnZ = 0.0
    n = 0
    while T.shape != torch.Size([args.chi, args.chi, args.chi, args.chi]): 
        #f = T.abs().max()
        #lnZ += 2**(-n)*torch.log(f)
        #T = T / f

        chi = args.D**(2**(n//2))
        U = torch.eye(chi**2, chi**2, device=device, dtype=dtype).view(chi, chi, chi**2)
        print (T.shape, U.shape)
        T = renormalize(T, U)
        n = n+1 

    print (T.shape, args.chi)

    isometries = hotrg(T)  
    hosrg(T, isometries)
